/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc:fr.bigeon.gclc.test.utils.WritingRunnable.java
 * Created on: Nov 29, 2016
 */
package fr.bigeon.gclc.manager;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/** A runnable to read the piped output.
 *
 * @author Emmanuel Bigeon */
public class ReadingRunnable implements Runnable {

    /** The closed pipe message */
    private static final String CLOSED_PIPE = "Closed pipe"; //$NON-NLS-1$
    /** Wait timeout */
    private static final long TIMEOUT = 1000;
    /** Class logger */
    private static final Logger LOGGER = Logger
            .getLogger(ReadingRunnable.class.getName());
    /** Read messages */
    private final Deque<String> messages = new ArrayDeque<>();
    /** the reader */
    private final BufferedReader reader;
    /** the state of this runnable */
    private boolean running = true;

    /** Synchro object */
    private final Object lock = new Object();
    /** The waiting status for a message */
    private boolean waiting;
    /**
     * The blocker for a given message
     */
    private final Map<String, Object> messageBlocker = new HashMap<>();
    /**
     * The lock
     */
    private final Object messageBlockerLock = new Object();
    private String delivering;

    /** @param reader the input to read from */
    public ReadingRunnable(BufferedReader reader) {
        super();
        this.reader = reader;
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run() */
    @Override
    public void run() {

        while (running) {
            try {
                String line = reader.readLine();
                if (line == null) {
                    // Buffer end
                    running = false;
                    return;
                }
                LOGGER.finer("Read: " + line); //$NON-NLS-1$
                line = stripNull(line);
                synchronized (lock) {
                    messages.add(line);
                    lock.notify();
                }
            } catch (InterruptedIOException e) {
                LOGGER.info("Reading interrupted"); //$NON-NLS-1$
                LOGGER.log(Level.FINER,
                        "Read interruption was caused by an exception", e); //$NON-NLS-1$
            } catch (IOException e) {
                LOGGER.severe("Unable to read from stream"); //$NON-NLS-1$
                LOGGER.log(Level.FINE, "The stream reading threw an exception", //$NON-NLS-1$
                        e);
                running = false;
                return;
            }
        }
    }

    /** Strip the string from head NULL characters.
     * 
     * @param line the line to strip the null character from
     * @return the resulting string */
    private static String stripNull(String line) {
        String res = line;
        while (res.length() > 0 && res.charAt(0) == 0) {
            LOGGER.severe(
                    "NULL character heading the result of the read. This is a stream problem..."); //$NON-NLS-1$
            res = res.substring(1);
        }
        return res;
    }

    /** @return the next read message
     * @throws IOException if the pipe is closed */
    public String getMessage() throws IOException {
        synchronized (lock) {
            if (!running) {
                throw new IOException(CLOSED_PIPE); 
            }
            waiting = true;
            while (messages.isEmpty()) {
                try {
                    lock.wait(TIMEOUT);
                } catch (InterruptedException e) {
                    LOGGER.log(Level.SEVERE, "Thread interruption exception.", //$NON-NLS-1$
                            e);
                }
                if (messages.isEmpty() && !running) {
                    throw new IOException(CLOSED_PIPE); 
                }
            }
            LOGGER.finest("Polled: " + messages.peek()); //$NON-NLS-1$
            waiting = false;
            notifyMessage(messages.peek());
            return messages.poll();
        }
    }

    /** @param running the running to set */
    public void setRunning(boolean running) {
        synchronized (lock) {
            this.running = running;
        }
    }

    /** @return the running */
    public boolean isRunning() {
        synchronized (lock) {
            return running;
        }
    }

    /** @return if a message is waiting
     * @throws IOException if the pipe is closed */
    public boolean hasMessage() throws IOException {
        synchronized (lock) {
            if (!running) {
                throw new IOException(CLOSED_PIPE); 
            }
            return !messages.isEmpty();
        }
    }

    /** Interrupts the wait on the next message by providing an empty message */
    public void interrupt() {
        synchronized (lock) {
            if (waiting) {
                messages.offer(""); //$NON-NLS-1$
                lock.notify();
            }
        }
    }

    /** @param message the message */
    private void notifyMessage(String message) {
        synchronized (messageBlockerLock) {
            delivering = message;
            if (messageBlocker.containsKey(message)) {
                Object mLock = messageBlocker.get(message);
                synchronized (mLock) {
                    mLock.notify();
                }
                messageBlocker.remove(message);
            }
        }
    }

    /** @param message the message
     * @return the thread to join to wait for message delivery */
    public Thread getWaitForDelivery(final String message) {
        synchronized (messageBlockerLock) {
            if (!messageBlocker.containsKey(message)) {
                messageBlocker.put(message, new Object());
            }
            final Object obj = messageBlocker.get(message);
            final Object start = new Object();
            Thread th = new Thread(new Runnable() {
    
                @SuppressWarnings("synthetic-access")
                @Override
                public void run() {
                    synchronized (obj) {
                        synchronized (start) {
                            start.notify();
                        }
                        while (isRunning()) {
                            try {
                                obj.wait();
                                if (delivering.equals(message)) {
                                    return;
                                }
                            } catch (InterruptedException e) {
                                LOGGER.log(Level.SEVERE,
                                        "Thread interruption exception.", e); //$NON-NLS-1$
                            }
                        }
                    }
                }
            });
            synchronized (start) {
                th.start();
                try {
                    start.wait();
                } catch (InterruptedException e) {
                    LOGGER.log(Level.SEVERE, "Thread interruption exception.", //$NON-NLS-1$
                            e);
                }
            }
            return th;
        }
    }
}
